import bw2calc
import bw2data
import bw2io
import numpy as np
import yaml

from unfold import Fold, Unfold


def test_db_reproduction():
    bw2data.projects.set_current("test")
    bw2io.create_default_biosphere3()

    fp = "./tests/fixture/reference_database.yaml"
    with open(fp, "r") as stream:
        db = yaml.load(stream, Loader=yaml.FullLoader)

    fp = "./tests/fixture/db A.yaml"

    with open(fp, "r") as stream:
        db_a = yaml.load(stream, Loader=yaml.FullLoader)

    fp = "./tests/fixture/db B.yaml"

    with open(fp, "r") as stream:
        db_b = yaml.load(stream, Loader=yaml.FullLoader)

    bw2data.Database("reference_database").write(db)
    bw2data.Database("db A").write(db_a)
    bw2data.Database("db B").write(db_b)

    lca = bw2calc.LCA({bw2data.get_activity(("db A", "activity A")): 1})
    lca.lci()
    original_supply_A = lca.supply_array

    lca = bw2calc.LCA({bw2data.get_activity(("db B", "activity A")): 1})
    lca.lci()
    original_supply_B = lca.supply_array

    f = Fold()

    f.fold(
        package_name="test",
        package_description="description of test",
        source="reference_database",
        system_model="cutoff",
        version="2.0",
        databases_to_fold=["db A", "db B"],
        descriptions=["this is db A", "this is db B"],
    )

    Unfold("test.zip").unfold(dependencies={"reference_database": "reference_database"})

    lca = bw2calc.LCA({bw2data.get_activity(("db A", "activity A")): 1})
    lca.lci()
    new_supply_A = lca.supply_array
    assert np.allclose(original_supply_A, new_supply_A) == True

    lca = bw2calc.LCA({bw2data.get_activity(("db B", "activity A")): 1})
    lca.lci()
    new_supply_B = lca.supply_array
    assert np.allclose(original_supply_B, new_supply_B) == True



import csv
import uuid
from pathlib import Path
from typing import List, Tuple, Union

import wurst as ws
import yaml
from prettytable import PrettyTable

DATA_DIR = Path(__file__).resolve().parent / "data"
OUTDATED_FLOWS = DATA_DIR / "outdated_flows.yaml"
OUTDATED_UNITS = DATA_DIR / "outdated_units.yaml"


def get_outdated_flows() -> dict:
    """
    Retrieve a list of outdated flows from the outdated flows file.
    """

    with open(OUTDATED_FLOWS, "r", encoding="utf-8") as stream:
        flows = yaml.safe_load(stream)

    return flows


def get_outdated_units() -> dict:
    """
    Retrieve a list of outdated flows from the outdated flows file.
    """

    with open(OUTDATED_UNITS, "r", encoding="utf-8") as stream:
        flows = yaml.safe_load(stream)

    return flows


def remove_missing_fields(data: List[dict]) -> List[dict]:
    """
    Remove any field that does not have information.
    """

    FORBIDDEN_FIELDS_TECH = [
        "categories",
    ]

    FORBIDDEN_FIELDS_BIO = [
        "location",
    ]

    def clean_up(exc: dict) -> dict:
        """Cleans up an exchange"""

        for field in list(exc.keys()):
            if exc[field] is None or exc[field] == "None":
                if field in exc:
                    del exc[field]
            if exc["type"] == "biosphere" and field in FORBIDDEN_FIELDS_BIO:
                if field in exc:
                    del exc[field]
            if exc["type"] == "technosphere" and field in FORBIDDEN_FIELDS_TECH:
                if field in exc:
                    del exc[field]

        return exc

    for dataset in data:
        for key, value in list(dataset.items()):
            if not value:
                del dataset[key]

        dataset["exchanges"] = [clean_up(exc) for exc in dataset["exchanges"]]

    return data


def check_exchanges_input(database, input_mapping):
    """
    Checks that all biosphere exchanges are given an input code

    """
    for dataset in database:
        for exc in dataset["exchanges"]:
            if exc["type"] == "biosphere":
                if "input" not in exc or exc.get("input") is None:
                    exc["input"] = input_mapping.get(
                        (
                            exc["name"],
                            exc.get("product"),
                            exc.get("location"),
                            exc.get("categories"),
                        )
                    )

    return database


def add_biosphere_links(
    data: List[dict], mapping: dict, delete_missing: bool = False
) -> List[dict]:
    """Add links for biosphere exchanges to :attr:`import_db`
    Modifies the :attr:`import_db` attribute in place.
    Also checks for outdated biosphere flows and replaces them with the
    current ones.

    :param data: list of dictionaries with the data to be imported
    :param delete_missing: whether unlinked exchanges should be deleted or not.
    """
    for x in data:
        for y in x["exchanges"]:
            if y["type"] == "biosphere":
                key = (
                    y["name"],
                    None,
                    None,
                    y["categories"],
                )
                if key in mapping:
                    y["input"] = (
                        "biosphere3",
                        mapping[key],
                    )
                else:
                    print(f"Could not find a biosphere flow for {key}")
        x["exchanges"] = [ex for ex in x["exchanges"] if "flag_deletion" not in ex]

    return data


def check_for_duplicates(db: List[dict], data: List[dict]) -> List[dict]:
    """
    Check whether the inventories to be imported are not
    already in the source database.
    :param db: list of dictionaries representing the source database
    :param data: list of dictionaries representing the inventories to be imported
    :returns: inventories without duplicates
    """

    # print if we find datasets that already exist
    db_names = [
        (x["name"].lower(), x["reference product"].lower(), x["location"]) for x in db
    ]

    already_exist = [
        (x["name"].lower(), x["reference product"].lower(), x["location"])
        for x in data
        if (x["name"].lower(), x["reference product"].lower(), x["location"])
        in db_names
    ]

    if len(already_exist) > 0:
        print(
            "The following datasets from the import file already exist in the source database:"
        )
        table = PrettyTable(["Name", "Reference product", "Location"])

        for dataset in already_exist:
            table.add_row([dataset[0][:50], dataset[1][:30], dataset[2]])

        print(table)
        print("They will not be imported.")

    return [
        x
        for x in data
        if (x["name"], x["reference product"], x["location"]) not in db_names
    ]


def add_product_field_to_exchanges(data: List[dict], db: List[dict]) -> List[dict]:
    """Add the `product` key to the production and
    technosphere exchanges in :attr:`import_db`.
    Also add `code` field if missing.
    For production exchanges, use the value of the `reference_product` field.
    For technosphere exchanges, search the activities in :attr:`import_db` and
    use the reference product. If none is found, search the Ecoinvent :attr:`database`.
    Modifies the :attr:`import_db` attribute in place.
    :param data: list of dictionaries with the data to be imported
    :raises IndexError: if no corresponding activity (and reference product) can be found.

    """
    # Add a `product` field to the production exchange
    for dataset in data:
        for exchange in dataset["exchanges"]:
            if exchange["type"] == "production":
                if "product" not in exchange:
                    exchange["product"] = dataset["reference product"]

                if exchange["name"] != dataset["name"]:
                    exchange["name"] = dataset["name"]

    # Add a `product` field to technosphere exchanges
    for dataset in data:
        for exchange in dataset["exchanges"]:
            if exchange["type"] == "technosphere":
                # Check if the field 'product' is present
                if "product" not in exchange:
                    try:
                        exchange["product"] = correct_product_field(exchange, data, db)
                    except KeyError:
                        print(
                            "No activity found for exchange {}{} in: {}".format(
                                exchange["name"], exchange["location"], dataset["name"]
                            )
                        )

    # Add a `code` field if missing
    for dataset in data:
        if "code" not in dataset:
            dataset["code"] = str(uuid.uuid4().hex)

    return data


def correct_product_field(exc: dict, data: List[dict], database: List[dict]) -> str:
    """
    Find the correct name for the `product` field of the exchange
    :param exc: a dataset exchange
    :return: name of the product field of the exchange

    """
    # Look first in the imported inventories
    candidate = next(
        ws.get_many(
            data,
            ws.equals("name", exc["name"]),
            ws.equals("location", exc["location"]),
            ws.equals("unit", exc["unit"]),
        ),
        None,
    )

    # If not, look in the ecoinvent inventories
    if candidate is None:
        candidate = next(
            ws.get_many(
                database,
                ws.equals("name", exc["name"]),
                ws.equals("location", exc["location"]),
                ws.equals("unit", exc["unit"]),
            ),
            None,
        )

    if candidate is not None:
        return candidate["reference product"]

    print(
        f"An inventory exchange in cannot be linked to the "
        f"biosphere or the ecoinvent database: {exc}"
    )

    return exc["reference product"]


def remove_categories_for_technosphere_flows(data):
    """
    Remove the categories field for technosphere flows
    :param data: list of dictionaries representing the inventories to be imported
    :returns: inventories
    """
    for x in data:
        for y in x["exchanges"]:
            if y["type"] in ["technosphere", "production"]:
                if "categories" in y:
                    del y["categories"]
    return data


def correct_fields_format(data: list, name: str) -> list:
    """
    Correct the format of some fields.
    :param data: database to check
    :return: database with corrected fields
    """

    for dataset in data:
        if "parameters" in dataset:
            if not isinstance(dataset.get("parameters", []), list):
                dataset["parameters"] = [dataset["parameters"]]

            if (
                dataset["parameters"] is None
                or dataset["parameters"] == {}
                or dataset["parameters"] == []
            ):
                del dataset["parameters"]

            else:
                new_parameters_list = []
                for p in dataset["parameters"]:
                    for k, v in p.items():
                        new_parameters_list.append({"name": k, "amount": v})
                dataset["parameters"] = new_parameters_list

        if "categories" in dataset:
            if dataset["categories"] is None:
                del dataset["categories"]
            elif not isinstance(dataset["categories"], tuple):
                dataset["categories"] = tuple(dataset["categories"])
            else:
                pass

        if not dataset.get("database"):
            dataset["database"] = name

    return data


def check_mandatory_fields(data: list) -> list:
    """
    Check that the mandatory fields are present.
    :param data: list of dictionaries representing the inventories to be imported
    :raises ValueError: if a mandatory field is missing
    :return: list of dictionaries representing the inventories to be imported
    """

    dataset_fields = [
        "name",
        "reference product",
        "location",
        "unit",
        "exchanges",
    ]

    missing_fields = []

    for dataset in data:
        for field in dataset_fields:
            if field not in dataset:
                if (
                    field in ["reference product", "location", "unit", "name"]
                    and "exchanges" in dataset
                ):
                    for exc in dataset["exchanges"]:
                        if exc["type"] == "production":
                            if field == "reference product":
                                dataset[field] = exc.get("product")
                            else:
                                dataset[field] = exc.get(field)

                if dataset.get(field) is None:
                    missing_fields.append(
                        [
                            dataset.get("database", "unknown"),
                            dataset.get("name", "unknown"),
                            dataset.get("reference product", "unknown"),
                            dataset.get("location", "unknown"),
                            field,
                        ]
                    )

    if missing_fields:
        # print in prettytable the list of missing fields
        table = PrettyTable()
        table.field_names = [
            "Database",
            "Dataset",
            "Reference product",
            "Location",
            "Missing field",
        ]
        for row in missing_fields[:10]:
            table.add_row(row)
        print(table)
        raise ValueError(
            "Some mandatory fields are missing in the database. "
            "Ten first missing fields are displayed above."
        )

    return data


def get_list_of_unique_datasets(data: list) -> list:
    """
    Return a list of unique datasets from a list of datasets.
    :param data: list of dictionaries representing the inventories to be imported
    :return: list of dictionaries representing the inventories to be imported
    """
    unique_datasets = []
    for dataset in data:
        if dataset not in unique_datasets:
            unique_datasets.append(
                (
                    dataset["name"],
                    dataset["reference product"],
                    dataset["location"],
                    dataset["unit"],
                )
            )
    return unique_datasets


def check_commonality_between_databases(original_db, scenario_db, db_name):
    """
    Check that scenario databases have at least one dataset in common
    with the original database.
    :param original_db: original
    :param scenario_db: list of scenario databases
    :param db_name: name of the scenario database
    :raises ValueError: if no dataset is in common
    """

    original_db_unique_datasets = get_list_of_unique_datasets(original_db)

    scenario_db_unique_datasets = get_list_of_unique_datasets(scenario_db)
    if not set(original_db_unique_datasets).intersection(scenario_db_unique_datasets):
        raise ValueError(
            "Could not find datasets in common between the reference database "
            f"and {db_name}."
        )


def change_db_name(data, name):
    """
    Originally from `wurst.linking` module.
    Change the database of all datasets in ``data`` to ``name``.

    Raises errors if each dataset does not have exactly one reference production exchange.
    """

    get_input_databases = lambda data: {ds.get("database") for ds in data}
    old_names = get_input_databases(data)
    for ds in data:
        ds["database"] = name
        for exc in ds["exchanges"]:
            if exc.get("input") and exc["input"][0] in old_names:
                exc["input"] = (name, exc["input"][1])
    return data



import itertools
from copy import copy

from bw2data import Database, databases
from bw2io.importers.base_lci import LCIImporter


def write_brightway_database(data, name):
    # Restore parameters to Brightway2 format
    # which allows for uncertainty and comments
    BW25UnfoldExporter(name, data).write_database()


class BW25UnfoldExporter(LCIImporter):
    """
    Inherits from `LCIImporter` to
    allow existing databases to be
    written to disk.

    """

    def __init__(self, db_name, data):
        super().__init__(db_name)
        self.db_name = db_name
        self.data = data
        for act in self.data:
            act["database"] = self.db_name

    # we override `write_database`
    # to allow existing databases
    # to be overwritten
    def write_database(self):
        def no_exchange_generator(data):
            for ds in data:
                cp = copy(ds)
                cp["exchanges"] = []
                yield cp

        if self.db_name in databases:
            print(
                f"Database {self.db_name} " f"already exists: it will be overwritten."
            )
        super().write_database(
            list(no_exchange_generator(self.data)), backend="iotable"
        )

        dependents = {exc["input"][0] for ds in self.data for exc in ds["exchanges"]}
        lookup = {
            obj.key: obj.id
            for obj in itertools.chain(*[Database(label) for label in dependents])
        }

        def technosphere_generator(data, lookup):
            for ds in data:
                target = lookup[(ds["database"], ds["code"])]
                for exc in ds["exchanges"]:
                    if exc["type"] in (
                        "substitution",
                        "production",
                        "generic production",
                    ):
                        yield {
                            "row": lookup[exc["input"]],
                            "col": target,
                            "amount": exc["amount"],
                            "flip": False,
                        }
                    elif exc["type"] == "technosphere":
                        yield {
                            "row": lookup[exc["input"]],
                            "col": target,
                            "amount": exc["amount"],
                            "flip": True,
                        }

        def biosphere_generator(data, lookup):
            for ds in data:
                target = lookup[(ds["database"], ds["code"])]
                for exc in ds["exchanges"]:
                    if exc["type"] == "biosphere":
                        yield {
                            "row": lookup[exc["input"]],
                            "col": target,
                            "amount": exc["amount"],
                            "flip": False,
                        }

        Database(self.db_name).write_exchanges(
            technosphere_generator(self.data, lookup),
            biosphere_generator(self.data, lookup),
            list(dependents),
        )



"""
Contains the Unfold class, to extract datapackage files.

"""

import copy
import os
import pickle
import uuid
from ast import literal_eval
from collections import defaultdict
from pathlib import Path
from typing import List, Union

import bw2data
import bw2io
import numpy as np
import pandas as pd
import sparse
from datapackage import Package
from prettytable import PrettyTable
from scipy import sparse as nsp
from wurst import extract_brightway2_databases
from wurst.linking import check_duplicate_codes, check_internal_linking, link_internal

from .data_cleaning import (
    add_biosphere_links,
    add_product_field_to_exchanges,
    change_db_name,
    check_exchanges_input,
    check_for_duplicates,
    correct_fields_format,
    get_outdated_flows,
    get_outdated_units,
    remove_categories_for_technosphere_flows,
    remove_missing_fields,
)

try:
    import bw_processing

    from .brightway25_export import write_brightway_database

except ImportError:
    from .brightway2_export import write_brightway_database

from .utils import HiddenPrints

DIR_CACHED_DB = Path(__file__).parent / "cached_databases"


def _c(value):
    """Converts zero to one."""
    if value == 0:
        return 1
    return value


def del_all(mapping, to_remove):
    """Remove list of elements from mapping."""
    for key in to_remove:
        del mapping[key]

    return mapping


def get_list_unique_exchanges(databases: list) -> list:
    """
    Gets a list of all unique exchanges found in a list of databases.

    :param databases: A list of Brightway2-style databases to extract unique exchanges from.
    :return: A list of tuples representing unique exchanges, where each tuple contains the following information:
        - name: The name of the exchange.
        - product: The reference product of the exchange.
        - categories: The categories of the exchange.
        - location: The location of the exchange.
        - unit: The unit of the exchange.
        - type: The type of the exchange (either "technosphere" or "biosphere").
    """

    return list(
        set(
            [
                (
                    exchange["name"],
                    exchange.get("product"),
                    exchange.get("categories"),
                    exchange.get("location"),
                    exchange.get("unit"),
                    exchange.get("type"),
                )
                for database in databases
                for dataset in database
                for exchange in dataset["exchanges"]
            ]
        )
    )


def check_cached_database(name) -> list:
    """
    Checks if a cached database exists, and if not, extracts it.
    :param name: name of the database
    :return: the database
    """

    # check that directory exists, otherwise create it
    Path(DIR_CACHED_DB).mkdir(parents=True, exist_ok=True)
    bw2io_version = "".join(tuple(map(str, bw2io.__version__)))
    file_name = Path(DIR_CACHED_DB / f"cached_{name}_{bw2io_version}.pickle")

    # check that file path leads to an existing file
    if file_name.exists() and len(bw2data.Database(name)) > 0:
        # check that the database has same length as the cached one
        database = pickle.load(open(file_name, "rb"))
        if len(database) == len(bw2data.Database(name)):
            # return the cached database
            return pickle.load(open(file_name, "rb"))

    # extract the database, pickle it for next time and return it
    print("Cannot find cached database. Will create one now for next time...")
    database = extract_brightway2_databases(name)
    pickle.dump(database, open(file_name, "wb"))
    return database


def clear_cache():
    """Clears the cache folder, except for files which contain __version__ in name.
    Useful when updating `premise`
    or encountering issues with
    inventories.
    """
    [f.unlink() for f in Path(DIR_CACHED_DB).glob("*")]


class Unfold:
    """Extracts datapackage files."""

    def __init__(self, path: Union[str, Path]):
        self.dict_meta = None
        self.acts_indices = None
        self.reversed_acts_indices = None
        self.outdated_flows = get_outdated_flows()
        self.outdated_units = get_outdated_units()
        self.path = path
        self.package = Package(self.path)
        self.dependencies = self.package.descriptor["dependencies"]
        self.scenarios = self.package.descriptor["scenarios"]
        self.scenario_df = None
        self.show_scenarios()
        self.database = []
        self.databases_to_export = {}
        self.dependency_mapping = {}
        self.factors = {}
        self.name = None

    def show_scenarios(self):
        """Shows the scenarios."""
        print("The data package contains the following scenarios:")
        table = PrettyTable()
        table.field_names = ["No.", "Scenario", "Description"]
        for i, scenario in enumerate(self.package.descriptor["scenarios"]):
            table.add_row([i, scenario["name"], scenario.get("description")])
        print(table)
        print("")
        print("To unfold all scenarios, `unfold()`.")
        print("To unfold a specific scenario, `unfold(scenarios=[1,])`.")

    def check_dependencies(self, dependencies: dict) -> None:
        """
        Checks the dependencies required to unfold the scenarios.

        Parameters:
        -----------
        dependencies : dict
            A dictionary containing the names of the databases required as dependencies.

        Returns:
        --------
        None

        Notes:
        ------
        This method checks that the required source database and scenarios are available and compatible with each other. If the dependencies are not provided, the method assumes that no dependencies are required.

        If the required dependencies are not found in the available databases, the method prompts the user to input the database number for each dependency.

        Once the user has provided the necessary information, the method updates the "source" attribute for each dependency with the name of the corresponding database.

        """
        # Get a list of available databases in the project
        available_databases = list(bw2data.databases.keys())

        # If no dependencies are provided, set an empty dict
        dependencies = dependencies or {}

        # Check if all dependencies are available in the project
        if dependencies and all(
            dependency in available_databases for dependency in dependencies.values()
        ):
            # If all dependencies are available, update the "source" attribute for each dependency
            for database in self.dependencies:
                database["source"] = dependencies[database["name"]]
        else:
            # If any dependencies are missing, prompt the user to input the necessary information
            print("The following databases are necessary to unfold the scenarios:")

            table = PrettyTable()
            table.field_names = ["No.", "Dependency", "System model", "Version"]

            # Add a row to the table for each dependency
            for _db, database in enumerate(self.dependencies):
                table.add_row(
                    [
                        _db + 1,
                        database["name"],
                        database.get("system model", ""),
                        database.get("version", ""),
                    ]
                )

            # Print the table of required dependencies
            print(table)
            print("")

            print("The following databases are available in your project:")

            table = PrettyTable()
            table.field_names = [
                "No.",
                "Database",
            ]

            # Add a row to the table for each available database
            for _db, database in enumerate(available_databases):
                table.add_row([_db + 1, database])

            # Print the table of available databases
            print(table)
            print("")

            # Prompt the user to input the database number for each dependency
            for _db, database in enumerate(self.dependencies):
                db_number = input(
                    f"Indicate the database number for dependency {_db + 1}: "
                )
                name = available_databases[int(db_number) - 1]
                database["source"] = name

    def build_mapping_for_dependencies(self, database) -> None:
        """
        Builds a mapping for dependencies based on the given database.

        Parameters:
        -----------
        database : list
            A list of activity dictionaries representing a database.

        Returns:
        --------
        None

        Notes:
        ------
        This method builds a mapping for dependencies based on the given database. The mapping is stored in the "dependency_mapping" attribute of the current object.

        The mapping is constructed using a dictionary comprehension, with each key being a tuple containing the name, reference product, location, and categories of the activity, and each value being a tuple containing the name and code of the database.

        """
        self.dependency_mapping.update(
            {
                (
                    a["name"],
                    a.get("reference product"),
                    a.get("location"),
                    a.get("categories"),
                ): (a["database"], a["code"])
                for a in database
            }
        )

    def extract_source_database(self):
        """
        Extracts the source database and builds a mapping for dependencies.

        Parameters:
        -----------
        None

        Returns:
        --------
        None

        Notes:
        ------
        This method extracts the source database for each dependency in the "dependencies" attribute of the current object, and builds a mapping for dependencies based on the extracted databases.
        The mapping is stored in the "dependency_mapping" attribute of the current object.

        If a dependency has a type of "source", the extracted database is also added to the "database" attribute of the current object.

        """

        for dependency in self.dependencies:
            database = check_cached_database(dependency["source"])

            self.build_mapping_for_dependencies(database)

            if dependency.get("type") == "source":
                self.database.extend(database)

    def clean_imported_inventory(self, data):
        """
        Cleans the imported inventory.

        Parameters:
        -----------
        data : list
            A list of activity dictionaries representing the imported inventory.

        Returns:
        --------
        list
            A cleaned list of activity dictionaries representing the imported inventory.

        Notes:
        ------
        This method cleans the imported inventory by performing several operations on it:

        - remove_missing_fields(): Removes fields that are missing or contain only None values.
        - add_biosphere_links(): Adds biosphere links to exchanges that are missing them.
        - check_for_duplicates(): Checks for duplicate activities and exchanges.
        - add_product_field_to_exchanges(): Adds a "product" field to each exchange.
        - remove_categories_for_technosphere_flows(): Removes categories for technosphere exchanges.

        The cleaned inventory is returned as a list of activity dictionaries.

        """

        print("Cleaning imported inventory...")
        data = remove_missing_fields(data)
        add_biosphere_links(data, self.dependency_mapping)
        check_for_duplicates(self.database, data)
        add_product_field_to_exchanges(data, self.database)
        remove_categories_for_technosphere_flows(data)
        return data

    def extract_additional_inventories(self) -> None:
        """Extracts additional inventories."""
        print("Extracting additional inventories...")
        with HiddenPrints():
            filepath = self.package.get_resource("inventories").source
            # ensure that all slashes in filepath are forward slashes
            filepath = filepath.replace("\\", "/")
            i = bw2io.CSVImporter(filepath)
            i.strategies = i.strategies[:4]
            i.apply_strategies()
            i.data = self.clean_imported_inventory(i.data)

        self.database.extend(i.data)
        self.database = change_db_name(self.database, self.package.descriptor["name"])
        self.build_mapping_for_dependencies(self.database)
        self.store_datasets_metadata()

    def generate_factors(self) -> None:
        """
        Calculates factors for each flow in the scenario based on
        the scenario_df dataframe.

        The resulting factors are stored in the factors attribute,
        which is a dictionary that maps flow ids to their corresponding
        factors.

        """
        self.factors = (
            self.scenario_df[
                [
                    c
                    for c in self.scenario_df.columns
                    if c
                    not in [
                        "to database",
                    ]
                ]
            ]
            .groupby("flow id")
            .sum(numeric_only=True)
            .to_dict("index")
        )

    def store_datasets_metadata(self) -> None:
        """
        Stores metadata for each dataset in the database attribute.

        The resulting metadata is stored in the dict_meta attribute, which is a dictionary that maps dataset identifiers to their
        metadata.

        """
        # store the metadata in a dictionary
        self.dict_meta = {
            (
                dataset["name"],
                dataset["reference product"],
                None,
                dataset["location"],
                dataset["unit"],
                "production",
            ): {
                key: values
                for key, values in dataset.items()
                if key
                not in [
                    "exchanges",
                    "code",
                    "name",
                    "reference product",
                    "location",
                    "unit",
                    "database",
                ]
            }
            for dataset in self.database
        }

    def generate_activities_indices(self) -> None:
        """Generates the indices of the unique activities."""
        # Get all unique activities from self.database.
        list_unique_acts = get_list_unique_exchanges(databases=[self.database])

        # Add additional exchanges to list_unique_acts.
        for act in list_unique_acts:
            if act[-1] == "production":
                new_id = list(act)
                new_id[-1] = "technosphere"
                new_id = tuple(new_id)
                if new_id not in list_unique_acts:
                    list_unique_acts.append(new_id)

        # Map each unique activity to a unique index.
        self.acts_indices = dict(enumerate(list_unique_acts))
        self.reversed_acts_indices = {act: i for i, act in enumerate(list_unique_acts)}

    def fetch_exchange_code(self, name: str, ref: str, loc: str) -> str:
        """
        :param name: name of the exchange
        :param ref: reference product of the exchange
        :param loc: location of the exchange
        :return: the exchange code corresponding to the specified name, reference product, and location
        Fetches the exchange code corresponding to the specified name,
        reference product, and location.
        """
        if (name, ref, loc, None) in self.dependency_mapping:
            return self.dependency_mapping[(name, ref, loc, None)][1]
        else:
            return str(uuid.uuid4().hex)

    def find_correct_id(self, key: tuple) -> [tuple, None]:
        if key[0] in self.outdated_flows:
            if (
                self.outdated_flows[key[0]],
                key[1],
                key[2],
                key[3],
            ) in self.dependency_mapping:
                return self.outdated_flows[key[0]], key[1], key[2], key[3]

            elif (
                self.outdated_flows[key[0]].replace(", ion", ""),
                key[1],
                key[2],
                key[3],
            ) in self.dependency_mapping:
                return (
                    self.outdated_flows[key[0]].replace(", ion", ""),
                    key[1],
                    key[2],
                    key[3],
                )

            elif (
                self.outdated_flows[key[0]] + ", ion",
                key[1],
                key[2],
                key[3],
            ) in self.dependency_mapping:
                return self.outdated_flows[key[0]] + ", ion", key[1], key[2], key[3]

            elif (
                key[0] + ", in ground",
                key[1],
                key[2],
                key[3],
            ) in self.dependency_mapping:
                return key[0] + ", in ground", key[1], key[2], key[3]

            else:
                print(f"Could not find key: {key}")
                # return key

    def fix_key(self, key: tuple) -> tuple:
        if key in self.dependency_mapping:
            return self.dependency_mapping[key]
        else:
            correct_id = self.find_correct_id(key)
            if correct_id is None:
                print(f"Could not find key: {key}")
                return key

            return self.dependency_mapping[correct_id]

    def get_exchange(
        self,
        ind: int,
        amount: float = 1.0,
        scenario_name: str = None,
    ):
        """
        Return an exchange in the form of a dictionary.
        If it has a value for "categories", we inder it is a biosphere flow.
        If not, it is either a technosphere or production flow.

        :param ind: index of the exchange
        :param amount: amount of the exchange
        :param scenario_name: name of the scenario
        :return: dictionary of the exchange
        """
        name, ref, cat, loc, unit, flow_type = self.acts_indices[ind]
        _ = lambda x: x if x != 0 else 1.0

        return {
            "name": name,
            "product": ref,
            "unit": unit,
            "location": loc,
            "categories": cat,
            "type": flow_type,
            "amount": amount if flow_type != "production" else _(amount),
            "input": (
                self.fix_key((name, ref, loc, cat))
                if flow_type == "biosphere"
                else (
                    scenario_name,
                    self.fetch_exchange_code(name, ref, loc),
                )
            ),
        }

    def populate_sparse_matrix(self) -> nsp.lil_matrix:
        """
        Generate a sparse matrix representation of the product system modeled by this object.

        The matrix is generated based on the data in the `database` attribute, which is assumed to be
        a list of dictionaries, where each dictionary represents a dataset in the product system.
        Each dataset has a list of exchanges, where each exchange is a dictionary with information
        about the input or output of a process.

        The matrix is generated as a `lil_matrix` object from the `scipy.sparse` module. The rows and
        columns of the matrix represent unique activities in the product system, which are determined
        by the unique combinations of the following exchange attributes: name, product, categories,
        location, unit, and type.

        The matrix is populated by looping over each exchange in each dataset, and adding the amount
        of the exchange to the corresponding row and column in the matrix.

        :return: A `lil_matrix` object representing the product system modeled by this object.
        """

        # Generate the indices for the activities
        self.generate_activities_indices()

        # Initialize a sparse matrix
        m = nsp.lil_matrix((len(self.acts_indices), len(self.acts_indices)))

        # Populate the matrix with exchange amounts
        for ds in self.database:
            for exc in ds["exchanges"]:
                # Source activity
                s = (
                    exc["name"],
                    exc.get("product"),
                    exc.get("categories"),
                    exc.get("location"),
                    exc["unit"],
                    exc["type"],
                )
                # Destination activity
                c = (
                    ds["name"],
                    ds.get("reference product"),
                    ds.get("categories"),
                    ds.get("location"),
                    ds["unit"],
                    "production",
                )
                # Add the exchange amount to the corresponding cell in the matrix
                m[self.reversed_acts_indices[s], self.reversed_acts_indices[c]] += exc[
                    "amount"
                ]

        return m

    def write_scaling_factors_in_matrix(
        self, matrix: np.ndarray, scenario_name: str
    ) -> np.ndarray:
        """
        Multiplies the elements of the given matrix with scaling factors for a given scenario.

        :param matrix: A 2D numpy array representing the matrix whose elements are to be scaled.
        :type matrix: numpy.ndarray

        :param scenario_name: A string representing the name of the scenario for which the scaling factors should be used.
        :type scenario_name: str

        :return: A 2D numpy array representing the scaled matrix, with the scaling factors applied to the appropriate elements.
        :rtype: numpy.ndarray
        """
        # Create a lambda function that returns 1.0 for zero values and the input value for non-zero values.
        # This is used to avoid multiplying by zero, which would result in a zero product.
        _ = lambda x: x if x != 0 else 1.0

        # Iterate over the scaling factors defined in `self.factors`.
        for flow_id, factor in self.factors.items():
            # Extract the components of the flow ID, which are used to look up the indices in the matrix.
            c_name, c_prod, c_loc, c_unit = list(flow_id)[:4]
            s_name, s_prod, s_loc, s_cat, s_unit, s_type = list(flow_id)[4:]

            # Look up the index of the consumer activity in the reversed activities index.
            consumer_idx = self.reversed_acts_indices.get(
                (
                    c_name,
                    c_prod,
                    None,
                    c_loc,
                    c_unit,
                    "production",
                ),
                None,
            )

            if consumer_idx is None:
                consumer_idx = self.reversed_acts_indices.get(
                    (
                        c_name,
                        c_prod,
                        None,
                        c_loc,
                        self.outdated_units.get(c_unit, c_unit),
                        "production",
                    ),
                    None,
                )

            # Look up the index of the supplier activity in the reversed activities index.
            supplier_id = (
                s_name,
                s_prod,
                s_cat,
                s_loc,
                s_unit,
                s_type,
            )
            supplier_idx = self.reversed_acts_indices.get(
                supplier_id,
                None,
            )

            if supplier_idx is None:
                supplier_idx = self.reversed_acts_indices.get(
                    (
                        s_name,
                        s_prod,
                        s_cat,
                        s_loc,
                        self.outdated_units.get(s_unit, s_unit),
                        s_type,
                    ),
                    None,
                )

            # Multiply the appropriate element of the matrix by the scaling factor for the given scenario.
            # Use the lambda function defined above to avoid multiplying by zero.
            if supplier_idx is not None and consumer_idx is not None:
                matrix[supplier_idx, consumer_idx] = factor[scenario_name] * _(
                    matrix[supplier_idx, consumer_idx]
                )
            else:
                print(
                    f"Could not find activity for flow {flow_id} in scenario {scenario_name}."
                )

        # Return the scaled matrix.
        return matrix

    def get_act_dict_structure(self, ind: int, scenario_name: str) -> dict:
        """
        Get the structure of the activity/dataset dictionary.
        :param ind: index of the activity
        :param scenario_name: name of the scenario
        :return: dictionary of the activity

        """
        name, ref, _, loc, unit, _ = self.acts_indices[ind]
        code = self.fetch_exchange_code(name, ref, loc)

        return {
            "name": name,
            "reference product": ref,
            "unit": unit,
            "location": loc,
            "database": scenario_name,
            "code": code,
            "parameters": [],
            "exchanges": [],
        }

    def build_superstructure_database(self, matrix: np.ndarray) -> list:
        """
        Builds a superstructure database from the given matrix, which is assumed to represent the exchanges between activities
        in a supply chain.

        :param matrix: A 3D numpy array representing the exchanges between activities in a supply chain.
        :type matrix: numpy.ndarray

        :return: A list of dictionaries representing the activities and exchanges in the supply chain.
        :rtype: list
        """
        # Print a message to indicate that the function is running.
        print(f"Generating superstructure database...")

        # Find the indices of the non-zero elements in the first matrix.
        non_zero_indices = sparse.argwhere(matrix[..., 0].T != 0)

        # Add the indices of the non-zero elements in the other matrices.
        for i in range(1, matrix.shape[-1]):
            non_zero_indices = np.concatenate(
                (non_zero_indices, sparse.argwhere(matrix[..., i].T != 0))
            )

        # Convert the indices to a list of tuples.
        non_zero_indices = list(map(tuple, non_zero_indices))

        # Group the indices by row in a dictionary.
        inds_d = defaultdict(list)
        for ind in non_zero_indices:
            if ind[1] not in inds_d[ind[0]]:
                inds_d[ind[0]].append(ind[1])

        # Build a list of activities and exchanges from the indices.
        new_db = []
        for k, v in inds_d.items():
            # Get the dictionary representing the current activity from the database.
            act = self.get_act_dict_structure(
                ind=k,
                scenario_name=self.name or self.package.descriptor["name"],
            )

            # Update the activity dictionary with metadata.
            act.update(self.dict_meta[self.acts_indices[k]])

            # Add exchanges to the activity dictionary for each non-zero element in the row.
            act["exchanges"].extend(
                self.get_exchange(
                    ind=j,
                    amount=matrix[j, k, 0],
                    scenario_name=self.name or self.package.descriptor["name"],
                )
                for j in v
            )

            # Add the updated activity dictionary to the list of activities and exchanges.
            new_db.append(act)

        # Return the list of activities and exchanges.
        return new_db

    def build_single_databases(
        self, matrix, databases_to_build: List[dict]
    ) -> list[list[dict]]:
        """
        Generate a list of single databases for each scenario specified in `databases_to_build`.

        :param matrix: A 3D numpy array representing the exchanges between activities in a supply chain.
        :type matrix: numpy.ndarray
        :param databases_to_build: A list of dictionaries, where each dictionary contains the name of a scenario and other scenario-specific metadata.
        :type databases_to_build: list[dict]
        :return: A list of single databases for each scenario specified in `databases_to_build`. Each database is a list of activity dictionaries.
        :rtype: list[list[dict]]

        :raises ValueError: If `databases_to_build` is empty.
        :raises ValueError: If the number of scenarios in `databases_to_build` does not match the number of scenarios in `matrix`.

        :notes:
        - The `matrix` parameter is a 3-dimensional array, with dimensions (n,m,p), where n and m are the number of rows
          and columns of the matrix and p is the number of scenarios.
        - The first two dimensions correspond to the producer and consumer indices of the exchanges, and the third dimension
          corresponds to the scenario index.
        - The matrix is assumed to be sparse, with zeros representing missing or invalid values.
        - Each database is represented as a list of activity dictionaries, where each activity dictionary contains the
          metadata and exchanges for a specific activity in the scenario.
        """

        databases_to_return = []

        for ix, i in enumerate(databases_to_build):
            # Print a message indicating that a database is being generated for the current scenario.
            print(f"Generating database for scenario {i['name']}...")

            # Get the indices of non-zero elements in the matrix for the current scenario.
            non_zero_indices = sparse.argwhere(matrix[..., ix].T != 0)
            non_zero_indices = list(map(tuple, non_zero_indices))

            # Create a dictionary that maps producer indices
            # to lists of consumer indices for the non-zero elements
            # in the matrix.
            inds_d = defaultdict(list)
            for ind in non_zero_indices:
                inds_d[ind[0]].append(ind[1])

            new_db = []

            # For each producer index, create an activity dictionary
            # and add it to the current scenario's database.
            for k, v in inds_d.items():
                act = self.get_act_dict_structure(
                    ind=k,
                    scenario_name=i["name"],
                )
                act.update(self.dict_meta[self.acts_indices[k]])

                # For each consumer index associated with the current producer index, create an exchange dictionary
                # and add it to the activity's exchanges list.
                act["exchanges"].extend(
                    self.get_exchange(
                        ind=j, amount=matrix[j, k, ix], scenario_name=i["name"]
                    )
                    for j in v
                )

                new_db.append(act)

            # remove datasets that are not in the current scenario
            new_db = self.filter_out_datasets(new_db, i["name"])

            # Add the current scenario's database
            # to the list of databases to return.
            databases_to_return.append(new_db)

        return databases_to_return

    def filter_out_datasets(
        self, database: List[dict], scenario_name: str
    ) -> List[dict]:
        """
        Remove datasets that are not in the current scenario.
        """

        # Get the list of datasets not used in the current scenario.
        df_gr = self.scenario_df.groupby(
            ["to activity name", "to reference product", "to location"]
        ).sum(numeric_only=True)

        datasets_not_in_scenario = df_gr.loc[
            (df_gr[scenario_name] == 0) & (df_gr.sum(1) != 0), :
        ].index.tolist()

        # remove items from datasets_not_in_scenario that are
        # used in exchanges of database
        for act in database:
            for exc in act["exchanges"]:
                if exc["type"] == "technosphere":
                    key = (exc["name"], exc["product"], exc["location"])
                    if key in datasets_not_in_scenario:
                        datasets_not_in_scenario.remove(key)

        # Remove datasets that are not in the current scenario.
        database = [
            act
            for act in database
            if (act["name"], act["reference product"], act["location"])
            not in datasets_not_in_scenario
        ]

        return database

    def check_usage(
        self, name: str, product: str, location: str, database: list
    ) -> bool:
        # check that name, product, location are not used in any exchange of database

        pass

    def generate_superstructure_database(self) -> List[dict]:
        """
        Generates the superstructure database.

        Returns:
        - A list of activity dictionaries representing the superstructure database.

        Notes:
        - This function first populates a sparse matrix with the exchanges between activities in the supply chain.
        - It then creates a 3D numpy array by stacking the sparse matrix with scaled versions of the matrix for each scenario in `self.scenarios`.
        - Finally, it uses the 3D numpy array to generate the superstructure database by calling the `build_superstructure_database()` function.
        """

        m = self.populate_sparse_matrix()

        matrix = sparse.stack(
            [sparse.COO(m)]
            + [
                sparse.COO(
                    self.write_scaling_factors_in_matrix(copy.deepcopy(m), s["name"])
                )
                for _, s in enumerate(self.scenarios)
            ],
            axis=-1,
        )

        return self.build_superstructure_database(matrix=matrix)

    def generate_single_databases(self) -> List[List[dict]]:
        """
        Generates single databases for each scenario in `self.scenarios`.

        Returns:
        - A list of lists of activity dictionaries representing the single databases.

        Notes:
        - This function first populates a sparse matrix with the exchanges between activities in the supply chain.
        - It then creates a 3D numpy array by stacking scaled versions of the matrix for each scenario in `self.scenarios`.
        - Finally, it uses the 3D numpy array to generate single databases for each scenario by calling the `build_single_databases()` function.
        """
        m = self.populate_sparse_matrix()

        matrix = sparse.stack(
            [
                sparse.COO(
                    self.write_scaling_factors_in_matrix(copy.deepcopy(m), s["name"])
                )
                for _, s in enumerate(self.scenarios)
            ],
            axis=-1,
        )

        return self.build_single_databases(
            matrix=matrix, databases_to_build=self.scenarios
        )

    def format_dataframe(
        self, scenarios: List[int] = None, superstructure: bool = False
    ):
        """
        Formats the scenario dataframe.

        :param scenarios: A list of scenario indices to keep in the dataframe. If `None`, keeps all scenarios.
        :type scenarios: list[int]
        :param superstructure: If `True`, adds columns to the dataframe for the superstructure.
        :type superstructure: bool

        .. note::
        - This function removes unused scenarios from `self.scenarios`, drops columns from the scenario dataframe that correspond to scenarios that were removed, and converts `None` and `np.nan` values to `None`.
        - If `superstructure` is `True`, this function adds columns to the scenario dataframe for the superstructure by calling the `get_suppliers()` and `get_consumers()` functions.

        """

        scenarios = scenarios or list(range(len(self.scenarios)))
        scenarios_to_keep = [self.scenarios[i]["name"] for i in scenarios]

        # Remove scenarios that are not in `scenarios_to_keep` from `self.scenarios`.
        self.scenarios = [s for s in self.scenarios if s["name"] in scenarios_to_keep]

        # Find the scenarios to leave out of the scenario dataframe.
        scenarios_to_leave_out = list(
            set(s["name"] for s in self.scenarios) - set(scenarios_to_keep)
        )

        # Load the scenario dataframe from the `scenario_data` resource.
        filepath = self.package.get_resource("scenario_data").source
        # ensure that all slashes in filepath are forward slashes
        filepath = filepath.replace("\\", "/")
        self.scenario_df = pd.read_csv(
            filepath, keep_default_na=False, na_values="", encoding="utf-8-sig"
        )

        # Drop columns corresponding to scenarios that were removed.
        self.scenario_df = self.scenario_df.drop(scenarios_to_leave_out, axis=1)

        # Convert "None" and np.nan values to None.
        self.scenario_df = self.scenario_df.replace("None", None)
        self.scenario_df = self.scenario_df.replace({np.nan: None})

        # Convert strings to Python objects in columns that contain lists or tuples.
        self.scenario_df["from categories"] = self.scenario_df["from categories"].apply(
            lambda x: literal_eval(str(x))
        )
        self.scenario_df["to categories"] = self.scenario_df["to categories"].apply(
            lambda x: literal_eval(str(x))
        )

        self.scenario_df["from key"] = self.scenario_df["from key"].apply(
            lambda x: literal_eval(str(x))
        )

        self.scenario_df["to key"] = self.scenario_df["to key"].apply(
            lambda x: literal_eval(str(x))
        )

        # Convert columns corresponding to scenario exchange amounts to float dtype.
        self.scenario_df[scenarios_to_keep] = self.scenario_df[
            scenarios_to_keep
        ].astype(float)

        # Add a "flow id" column to the scenario dataframe.
        self.scenario_df["flow id"] = list(
            zip(
                self.scenario_df["to activity name"],
                self.scenario_df["to reference product"],
                self.scenario_df["to location"],
                self.scenario_df["to unit"],
                self.scenario_df["from activity name"],
                self.scenario_df["from reference product"],
                self.scenario_df["from location"],
                self.scenario_df["from categories"],
                self.scenario_df["from unit"],
                self.scenario_df["flow type"],
            )
        )

    def format_superstructure_dataframe(self) -> None:
        """
        Formats the superstructure dataframe.

        Notes:
        - This function creates a new scenario dataframe for the superstructure by populating a sparse matrix with data from the self.factors dictionary.
        - The superstructure dataframe contains columns for each scenario as well as additional columns for database information and metadata.
        """

        # Create a sparse matrix with data from self.factors
        matrix = self.populate_sparse_matrix()

        # Define a lambda function to replace zero values with 1.0
        _ = lambda x: 1.0 if x == 0.0 else x

        # Loop through each flow in self.factors and update the corresponding factor values
        for flow_id, factor in self.factors.items():
            c_name, c_prod, c_loc, c_unit = list(flow_id)[:4]
            s_name, s_prod, s_loc, s_cat, s_unit, s_type = list(flow_id)[4:]

            # Get the indices of the consumer and supplier activities
            # in the reversed_act_indices dictionary
            consumer_key = (
                c_name,
                c_prod,
                None,
                c_loc,
                c_unit,
                "production",
            )
            if consumer_key in self.reversed_acts_indices:
                consumer_idx = self.reversed_acts_indices[consumer_key]
            else:
                consumer_key = (
                    c_name,
                    c_prod,
                    None,
                    c_loc,
                    self.outdated_units.get(c_unit, c_unit),
                    "production",
                )

                if consumer_key in self.reversed_acts_indices:
                    consumer_idx = self.reversed_acts_indices[consumer_key]
                else:
                    print("not found key for consumer ", consumer_key)
                    consumer_idx = 0

            supplier_key = (
                s_name,
                s_prod,
                s_cat,
                s_loc,
                s_unit,
                s_type,
            )

            if supplier_key not in self.reversed_acts_indices:
                try:
                    (
                        s_name,
                        s_prod,
                        s_loc,
                        s_cat,
                    ) = self.find_correct_id((s_name, s_prod, s_loc, s_cat))

                    supplier_key = (
                        s_name,
                        s_prod,
                        s_cat,
                        s_loc,
                        s_unit,
                        s_type,
                    )
                except TypeError:
                    supplier_key = (
                        s_name,
                        s_prod,
                        s_cat,
                        s_loc,
                        self.outdated_units.get(s_unit, s_unit),
                        s_type,
                    )

                    if supplier_key not in self.reversed_acts_indices:
                        print("not found key for supplier", supplier_key)

            supplier_idx = self.reversed_acts_indices[supplier_key]

            # Update the factor values for each scenario
            # by multiplying with the corresponding matrix value

            for scenario, val in factor.items():
                factor[scenario] = val * _(matrix[supplier_idx, consumer_idx])

        # Create a new scenario dataframe from the updated factors dictionary
        self.scenario_df = pd.DataFrame.from_dict(self.factors).T.reset_index()

        # Rename columns
        # and add new columns
        # for database information and metadata

        self.scenario_df.columns = [
            "to activity name",
            "to reference product",
            "to location",
            "to unit",
            "from activity name",
            "from reference product",
            "from location",
            "from categories",
            "from unit",
            "flow type",
        ] + [s["name"] for s in self.scenarios]

        self.scenario_df["to database"] = self.name or self.package.descriptor["name"]
        self.scenario_df["to categories"] = None
        self.scenario_df["to key"] = None
        self.scenario_df["from key"] = None
        self.scenario_df = self.scenario_df.replace({np.nan: None})

        # if self.name is not None, we change the name of the database
        # in self.dependency_mapping to the name of the superstructure
        if self.name:
            self.dependency_mapping = {
                k: (self.name, v[1]) if v[0] == self.package.descriptor["name"] else v
                for k, v in self.dependency_mapping.items()
            }

        # Use the dependency_mapping dictionary
        # to add "from key" and "to key" columns
        self.scenario_df.loc[:, "from key"] = self.scenario_df.apply(
            lambda x: self.dependency_mapping.get(
                (
                    x["from activity name"],
                    x["from reference product"],
                    x["from location"],
                    x["from categories"],
                )
            ),
            axis=1,
        )
        self.scenario_df.loc[:, "to key"] = self.scenario_df.apply(
            lambda x: self.dependency_mapping.get(
                (
                    x["to activity name"],
                    x["to reference product"],
                    x["to location"],
                    x["to categories"],
                )
            ),
            axis=1,
        )

        # Add "from database" column based on flow type
        # set dtype of "from database" and "to database" to string

        # create "from database" column
        self.scenario_df.loc[:, "from database"] = None

        self.scenario_df = self.scenario_df.astype(
            {"from database": "string"},
            {"to database": "string"},
        )

        self.scenario_df.loc[
            (self.scenario_df["flow type"] == "technosphere")
            | (self.scenario_df["flow type"] == "production"),
            "from database",
        ] = (
            self.name or self.package.descriptor["name"]
        )

        self.scenario_df.loc[
            (self.scenario_df["flow type"] == "biosphere"), "from database"
        ] = "biosphere3"
        self.scenario_df = self.scenario_df[
            [
                "from activity name",
                "from reference product",
                "from location",
                "from categories",
                "from database",
                "from key",
                "to activity name",
                "to reference product",
                "to location",
                "to categories",
                "to database",
                "to key",
                "flow type",
            ]
            + [s["name"] for s in self.scenarios]
        ]

    def unfold(
        self,
        scenarios: List[int] = None,
        dependencies: dict = None,
        superstructure: bool = False,
        export_dir: str = None,
        name: str = None,
    ):
        """
        unfold() is a method of the Unfold class, which extracts specific scenarios from the input LCA database and writes them as new databases.

        :param scenarios: A list of integers indicating the indices of the scenarios to extract. If None, all scenarios are extracted. Default is None.
        :param dependencies: A dictionary containing additional inventory databases that may be needed for extraction. Default is None.
        :param superstructure: A boolean indicating whether to generate a scenario difference file and a superstructure database. Default is False.
        :param name: A string indicating the name of the superstructure database. Default is None.
        :param export_dir: A string indicating the directory to which the scenario difference file should be exported. Default is current working directory.
        :return: None

        Behavior:

        Calls the check_dependencies() method to ensure that all required inventory databases are present.
        Calls the extract_source_database() method to extract the original LCA database.
        Calls the extract_additional_inventories() method to extract any additional inventory databases required.
        Calls the format_dataframe() method to format the scenario data into a Pandas DataFrame.
        Calls the generate_factors() method to generate the factors for the scenarios.
        If superstructure is False, calls the generate_single_databases() method to generate a separate LCA database for each scenario, and stores them in the databases_to_export dictionary.
        If superstructure is True, calls the format_superstructure_dataframe() method to generate a scenario difference file and a superstructure database.
        Calls write() method to write the databases to disk.

        """

        self.name = name
        self.check_dependencies(dependencies)
        self.extract_source_database()
        self.extract_additional_inventories()
        self.format_dataframe(scenarios=scenarios, superstructure=superstructure)
        self.generate_factors()

        if not superstructure:
            self.databases_to_export = {
                k: v
                for k, v in zip(
                    [s["name"] for s in self.scenarios],
                    self.generate_single_databases(),
                )
            }
        else:
            print("Writing scenario difference file...")
            self.format_superstructure_dataframe()
            self.database = self.generate_superstructure_database()

        self.write(superstructure=superstructure, export_dir=export_dir)

    def write(self, superstructure: bool = False, export_dir: str = None):
        """
        Write the databases.
        If superstructure is True, write the scenario difference file,
        along with a database.

        :param superstructure: bool, default False
        :param export_dir: str, default current working directory

        """

        if not superstructure:
            for scenario, database in self.databases_to_export.items():
                change_db_name(data=database, name=scenario)
                link_internal(database)
                check_internal_linking(database)
                check_duplicate_codes(database)
                correct_fields_format(database, scenario)
                print(f"Writing database for scenario {scenario}...")
                write_brightway_database(data=database, name=scenario)

        else:
            source_db = [db for db in self.dependencies if db.get("type") == "source"]

            if len(source_db) == 1:
                source_db = source_db[0]
            else:
                source_db = {"name": "unknown", "version": "unknown"}

            if self.name:
                filename = f"SDF {source_db['name']} {self.name}.csv"
            else:
                filename = f"SDF {source_db['name']} {source_db['version']} {self.package.descriptor['name']}.csv"

            if export_dir is None:
                export_dir = os.getcwd()

            filename = os.path.join(export_dir, filename)

            self.scenario_df.to_csv(
                filename,
                index=False,
                encoding="utf-8-sig",
            )

            print(f"Scenario difference file exported to {filename}.")
            print("")
            print("Writing superstructure database...")
            change_db_name(self.database, self.name or self.package.descriptor["name"])
            self.database = check_exchanges_input(
                self.database, self.dependency_mapping
            )
            link_internal(self.database)
            check_internal_linking(self.database)
            check_duplicate_codes(self.database)
            correct_fields_format(
                self.database, self.name or self.package.descriptor["name"]
            )
            write_brightway_database(
                data=self.database, name=self.name or self.package.descriptor["name"]
            )



from bw2data import databases
from bw2io.importers.base_lci import LCIImporter


def write_brightway_database(data, name):
    # Restore parameters to Brightway2 format
    # which allows for uncertainty and comments
    BW2UnfoldExporter(name, data).write_database()


class BW2UnfoldExporter(LCIImporter):
    """
    Inherits from `LCIImporter` to
    allow existing databases to be
    written to disk.

    """

    def __init__(self, db_name, data):
        super().__init__(db_name)
        self.db_name = db_name
        self.data = data
        for act in self.data:
            act["database"] = self.db_name

    # we override `write_database`
    # to allow existing databases
    # to be overwritten
    def write_database(self):
        if self.db_name in databases:
            print(
                f"Database {self.db_name} " f"already exists: it will be overwritten."
            )
            del databases[self.db_name]
        super().write_database()



import os
import sys


class HiddenPrints:
    """
    From https://stackoverflow.com/questions/8391411/how-to-block-calls-to-print
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w", encoding="utf-8")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout



"""
Contains the Fold class, which is used to fold one
or several databases into a data package.

"""

import csv
import uuid
from collections import defaultdict
from functools import lru_cache
from pathlib import Path
from typing import List, Union

import bw2data
import pandas as pd
import sparse
from datapackage import Package
from prettytable import PrettyTable
from scipy import sparse as nsp

from . import __version__
from .data_cleaning import (
    DATA_DIR,
    check_commonality_between_databases,
    check_mandatory_fields,
)

DIR_DATAPACKAGE_TEMP = DATA_DIR / "temp"


def write_formatted_data(name, data, filepath):
    """
    Adapted from bw2io.export.csv
    :param name: name of the database
    :param data: data to write
    :param filepath: path to the file
    """

    sections = [
        "project parameters",
        "database",
        "database parameters",
        "activities",
        "activity parameters",
        "exchanges",
    ]

    result = []

    if "database" in sections:
        result.append(["Database", name])
        result.append([])

    if "activities" not in sections:
        return result

    for act in data:
        result.append(["Activity", act["name"]])
        result.append(["reference product", act["reference product"]])
        result.append(["unit", act["unit"]])
        result.append(["location", act["location"]])
        result.append(["comment", act.get("comment", "")])
        result.append(["source", act.get("source", "")])
        result.append([""])

        if "exchanges" in sections:
            result.append(["Exchanges"])
            if act.get("exchanges"):
                result.append(
                    [
                        "name",
                        "amount",
                        "unit",
                        "location",
                        "categories",
                        "type",
                        "product",
                    ]
                )
                for exc in act["exchanges"]:
                    result.append(
                        [
                            exc["name"],
                            exc["amount"],
                            exc["unit"],
                            exc.get("location"),
                            (
                                "::".join(list(exc.get("categories", [])))
                                if exc["type"] == "biosphere"
                                else None
                            ),
                            exc["type"],
                            exc.get("product"),
                        ]
                    )
        result.append([])

    with open(filepath, "w", newline="", encoding="utf-8") as file:
        writer = csv.writer(file)
        for line in result:
            writer.writerow(line)

    return filepath


class Fold:
    """
    Folds a list of databases into a data package.
    """

    def __init__(self):
        self.exc_codes = {}
        self.source = None
        self.databases_to_fold = []
        self.dependencies = set()
        self.dependency_mapping = {}
        self.rev_dependency_mapping = {}
        self.datapackage_name = None
        self.datapackage_description = None

    @staticmethod
    def extract_database(db_name):
        """Extracts the source database."""
        return [v for v in bw2data.Database(db_name).load().values()]

    def identify_databases_to_fold(
        self,
        source_database: str = None,
        source_database_system_model: str = None,
        source_database_version: Union[float, str] = None,
        databases_to_fold: List[str] = None,
        descriptions: List[str] = None,
    ):
        """
        The identify_databases_to_fold function identifies the source database and the databases to be folded into the source database and extracts them.

        :param source_database: Name of the source database to be used. If not specified, the user is prompted to choose from the available databases.
        :param source_database_system_model: System model of the source database to be used. If not specified, the user is prompted to input.
        :param source_database_version: Version of the source database to be used. If not specified, the user is prompted to input.
        :param databases_to_fold: List of databases to be folded into the source database. If not specified, the user is prompted to input.
        :param descriptions: Short descriptions of each database to be folded. If not specified, the user is prompted to input.
        :return: The source dictionary containing information about the source database, such as its name, database, system model, and version.
        :return: The databases_to_fold list containing dictionaries of information about each database to be folded, including its name, database, and description.

        Functionality:

        Checks whether the user has already specified a datapackage name and description. If not, prompts the user to input these details.
        Lists the available databases and prompts the user to input the number of the reference database if the source_database input is not specified.
        Identifies the dependencies of the source and folded databases.
        Prompts the user to input the system model and version of the source database if not specified.
        Prompts the user to input the list of databases to be folded and their descriptions if not specified.
        Extracts the source database and ensures that mandatory fields are included.
        Builds the mapping of dependencies for the source database.
        Extracts each database to be folded, ensures that mandatory fields are included, and builds the mapping of dependencies.
        Identifies whether any dependencies are external and, if so, extracts them.
        Returns a set of the dependencies excluding the source and databases to be folded.

        """

        if not self.datapackage_name:
            self.datapackage_name = str(input("Give a name for this datapackage: "))

        if not self.datapackage_description:
            self.datapackage_description = str(
                input("Give a short description for this datapackage: ")
            )

        available_databases = list(bw2data.databases.keys())

        if not source_database:
            print("The following databases are available in your project:")

            table = PrettyTable()
            table.field_names = [
                "No.",
                "Database",
            ]

            for _db, database in enumerate(available_databases):
                table.add_row([_db + 1, database])
            print(table)
            print("")

            source_database = int(input("Indicate the no. of the reference database: "))
            source_database = available_databases[source_database - 1]
        assert source_database in available_databases, "Source database not found"

        self.identify_dependencies(source_database)

        if not source_database_system_model:
            source_database_system_model = str(
                input(f"Indicate the system model of {source_database}: ")
            )
        if not source_database_version:
            source_database_version = str(
                input(f"Indicate the version of {source_database}: ")
            )

        if not databases_to_fold:
            databases_to_fold = str(
                input(
                    "Indicate the no. of the databases "
                    "you would like to fold, "
                    "separated by a comma: "
                )
            )
            databases_to_fold = [
                available_databases[int(db.strip()) - 1]
                for db in databases_to_fold.split(",")
            ]

        if not descriptions:
            databases_descriptions = {
                db: str(input(f"Give a short description for {db}: "))
                for db in databases_to_fold
            }
        else:
            databases_descriptions = dict(zip(databases_to_fold, descriptions))

        assert all(
            database in available_databases for database in databases_to_fold
        ), "Database not found"

        self.identify_dependencies(databases_to_fold)

        print("Extracting source database...")
        source_database_extracted = self.extract_database(source_database)
        self.source = {
            "name": source_database,
            "database": source_database_extracted,
            "system model": source_database_system_model,
            "version": source_database_version,
        }

        source_database_extracted = check_mandatory_fields(source_database_extracted)

        self.build_mapping_for_dependencies(source_database_extracted)
        print("Done!")

        print("Extracting databases to fold...")
        for database in databases_to_fold:
            extracted_database = self.extract_database(database)
            extracted_database = check_mandatory_fields(extracted_database)
            check_commonality_between_databases(
                source_database_extracted, extracted_database, database
            )

            self.databases_to_fold.append(
                {
                    "name": database,
                    "database": extracted_database,
                    "description": databases_descriptions[database],
                }
            )
            self.build_mapping_for_dependencies(extracted_database)

        print("Done!")

        for dependency in self.dependencies:
            if dependency not in [source_database] + databases_to_fold:
                if dependency in bw2data.databases:
                    print("")
                    print(
                        f"The database {dependency} is an external dependency. "
                        f"It will be extracted automatically."
                    )
                    self.build_mapping_for_dependencies(
                        self.extract_database(dependency)
                    )
                else:
                    raise ValueError(
                        f"Database {dependency} also needed but not found."
                    )

        self.dependencies = {
            dependency
            for dependency in self.dependencies
            if dependency not in [source_database] + databases_to_fold
        }

    def build_mapping_for_dependencies(self, database):
        """Builds a mapping for dependencies."""
        self.dependency_mapping.update(
            {
                (
                    dataset["name"],
                    dataset.get("reference product"),
                    dataset.get("location"),
                    dataset.get("unit"),
                    dataset.get("categories"),
                ): (dataset["database"], dataset["code"])
                for dataset in database
            }
        )

        self.rev_dependency_mapping.update(
            {
                (dataset["database"], dataset["code"]): (
                    dataset["name"],
                    dataset.get("reference product"),
                    dataset.get("location"),
                    dataset.get("unit"),
                    dataset.get("categories"),
                )
                for dataset in database
            }
        )

    def get_list_unique_acts(self, scenarios: List[List[dict]]) -> list:
        """
        Returns a list of unique activities from a list of databases, where each database is represented by a list of
        datasets containing their respective exchanges.

        :param scenarios: A list of databases, where each database is a list of datasets, with each dataset containing the
                          exchanges.
        :type scenarios: list

        :return: A list of tuples representing the unique activities in the provided databases, where each tuple contains
                 the activity name, reference product, location, categories, unit and type.
        :rtype: list
        """

        list_unique_acts = []
        for database in scenarios:
            for dataset in database:
                list_unique_acts.extend(
                    [
                        (
                            self.rev_dependency_mapping.get(
                                exchange.get("input"), (None,)
                            )[0],
                            self.rev_dependency_mapping.get(
                                exchange.get("input"), (None, None)
                            )[1],
                            self.rev_dependency_mapping.get(
                                exchange.get("input"), (None,)
                            )[-1],
                            self.rev_dependency_mapping.get(
                                exchange.get("input"), (None, None, None)
                            )[2],
                            exchange["unit"],
                            exchange["type"],
                        )
                        for exchange in dataset["exchanges"]
                    ]
                )
        return list(set(list_unique_acts))

    def identify_dependencies(self, database_names):
        """
        Identify the dependencies between the databases
        :return: dictionary of dependencies
        """

        if isinstance(database_names, str):
            database_names = [database_names]

        for name in database_names:
            self.dependencies.update(bw2data.Database(name).find_graph_dependents())

    def fold(
        self,
        package_name: str = None,
        package_description: str = None,
        source: str = None,
        system_model: str = None,
        version: Union[float, str] = None,
        databases_to_fold: List[str] = None,
        descriptions: List[str] = None,
    ):
        """
        Folds one or more databases into a new package.

        :param package_name: Name for the new datapackage.
        :type package_name: str, optional
        :param package_description: Short description for the new datapackage.
        :type package_description: str, optional
        :param source: Name of the source database.
        :type source: str, optional
        :param system_model: System model of the source database.
        :type system_model: str, optional
        :param version: Version of the source database.
        :type version: float or str, optional
        :param databases_to_fold: List of names of the databases to fold.
        :type databases_to_fold: List[str], optional
        :param descriptions: Short description for each database to fold.
        :type descriptions: List[str], optional
        :raises AssertionError: When one or more databases to fold are not found.
        """
        self.datapackage_name = package_name
        self.datapackage_description = package_description

        self.identify_databases_to_fold(
            source_database=source,
            source_database_system_model=system_model,
            source_database_version=version,
            databases_to_fold=databases_to_fold,
            descriptions=descriptions,
        )
        dataframe, extra_inventories = self.generate_scenario_factor_file(
            origin_db=self.source, scenarios=self.databases_to_fold
        )

        extra_inventories = self.zero_out_exchanges(extra_inventories)
        self.build_datapackage(dataframe, extra_inventories)

    def zero_out_exchanges(self, extra_inventories: List[dict]) -> List[dict]:
        """
        Zero out exchanges (except production exchanges)
        that are not in the source database.

        :param extra_inventories: list of inventories that are not in the source database
        :return: list of inventories with zeroed out exchanges
        """
        for dataset in extra_inventories:
            for exchange in dataset["exchanges"]:
                if exchange["type"] != "production":
                    exchange["amount"] = 0

        return extra_inventories

    def get_exchange(
        self, ind: int, acts_ind: dict, amount: float = 1.0, production: bool = False
    ):
        """
        Return an exchange in teh form of a dictionary.
        If it has a value for "categories", we inder it is a biosphere flow.
        If not, it is either a technosphere or production flow.

        :param ind: index of the exchange
        :param acts_ind: dictionary of activities
        :param amount: amount of the exchange
        :param production: boolean indicating if it is a production flow
        :return: dictionary of the exchange
        """
        name, ref, cat, loc, unit, flow_type = acts_ind[ind]
        _ = lambda x: x if x != 0 else 1.0
        return {
            "name": name,
            "product": ref,
            "unit": unit,
            "location": loc,
            "categories": cat,
            "type": flow_type,
            "amount": amount if flow_type != "production" else _(amount),
            "input": (
                self.dependency_mapping[(name, ref, loc, unit, cat)]
                if flow_type == "biosphere"
                else (
                    self.datapackage_name,
                    self.fetch_exchange_code(name, ref, loc, unit),
                )
            ),
        }

    @lru_cache()
    def fetch_exchange_code(self, name: str, ref: str, loc: str, unit: str) -> str:
        """
        Fetch the code of an activity or create one.
        name: name of the activity
        ref: reference product
        loc: location
        unit: unit

        :return: code of the activity
        """

        if (name, ref, loc, unit) not in self.exc_codes:
            code = str(uuid.uuid4().hex)
            self.exc_codes[(name, ref, loc, unit)] = code
        else:
            code = self.exc_codes[(name, ref, loc, unit)]

        return code

    def get_act_dict_structure(self, ind: int, acts_ind: dict) -> dict:
        """
        Get the structure of the activity/dataset dictionary.
        :param ind: index of the activity
        :param acts_ind: dictionary of activities
        :return: dictionary of the activity

        """
        name, ref, _, loc, unit, _ = acts_ind[ind]
        code = self.fetch_exchange_code(name, ref, loc, unit)

        return {
            "name": name,
            "reference product": ref,
            "unit": unit,
            "location": loc,
            "database": self.datapackage_name,
            "code": code,
            "parameters": [],
            "exchanges": [],
        }

    def check_for_outdated_flows(self, database):
        """
        Check for outdated flows in the database.
        The list of outdated flows is stored in the attribute outdated_flows,
        which is a dictionary with the old flow name as key and the new flow name as value.
        See outdated_flows.yaml for the list of outdated flows.

        :param database: the database to check
        :return: a database with the outdated flows replaced
        """

        for dataset in database:
            for exchange in dataset["exchanges"]:
                if exchange["name"] in self.outdated_flows:
                    exchange["name"] = self.outdated_flows[exchange["name"]]
        return database

    def build_datapackage(self, dataframe, inventories):
        """
        Create and export a scenario datapackage.
        """

        # check that directory exists, otherwise create it
        Path(DIR_DATAPACKAGE_TEMP).mkdir(parents=True, exist_ok=True)
        dataframe.to_csv(DIR_DATAPACKAGE_TEMP / "scenario_data.csv", index=False)
        write_formatted_data(
            name=self.datapackage_name,
            data=inventories,
            filepath=DIR_DATAPACKAGE_TEMP / "inventories.csv",
        )
        package = Package(base_path=str(DIR_DATAPACKAGE_TEMP))
        package.infer("**/*.csv")
        package.descriptor["name"] = self.datapackage_name
        package.descriptor["title"] = self.datapackage_name.capitalize()
        package.descriptor["description"] = self.datapackage_description
        package.descriptor["unfold version"] = str(__version__)
        package.descriptor["dependencies"] = [
            {
                "name": self.source["name"],
                "system model": self.source["system model"],
                "version": self.source["version"],
                "type": "source",
            }
        ]
        for dependency in self.dependencies:
            package.descriptor["dependencies"].append(
                {
                    "name": dependency,
                    "type": "dependency",
                }
            )
        package.descriptor["scenarios"] = [
            {"name": s["name"], "description": s["description"]}
            for s in self.databases_to_fold
        ]
        package.descriptor["licenses"] = [
            {
                "id": "CC0-1.0",
                "title": "CC0 1.0",
                "url": "https://creativecommons.org/publicdomain/zero/1.0/",
            }
        ]
        package.commit()
        package.save(f"{self.datapackage_name}.zip")

        print(f"Data package saved at {f'{self.datapackage_name}.zip'}")

    def generate_scenario_factor_file(self, origin_db, scenarios):
        """
        Generate a scenario factor file from a list of databases
        :param origin_db: the original database
        :param scenarios: a list of databases
        """

        print("Building scenario factor file...")

        # create the dataframe
        dataframe, new_db, list_unique_acts = self.generate_scenario_difference_file(
            origin_db=origin_db, scenarios=scenarios
        )

        original = dataframe["original"]
        original = original.replace(0, 1)
        dataframe.loc[:, "original":] = dataframe.loc[:, "original":].div(
            original, axis=0
        )

        # remove the column `original`
        dataframe = dataframe.drop(columns=["original"])
        # fetch a list of activities not present in original_db
        list_original_acts = self.get_list_unique_acts([origin_db["database"]])
        new_acts_list = list(set(list_unique_acts) - set(list_original_acts))

        # fetch the additional activities from new_db
        extra_acts = [
            dataset
            for dataset in new_db
            if (
                dataset["name"],
                dataset.get("reference product"),
                None,
                dataset.get("location"),
                dataset["unit"],
                "production",
            )
            in new_acts_list
        ]

        return dataframe, extra_acts

    def generate_scenario_difference_file(
        self, origin_db: dict, scenarios: List[dict]
    ) -> tuple[pd.DataFrame, list[dict], list[tuple]]:
        """
        Generate a scenario difference file for a given list of databases.
        The function generate_scenario_difference_file calculates the scenario difference file for a given list of databases.
        This function takes in two parameters, origin_db, and scenarios. origin_db is a dictionary representing the
        original database, and scenarios is a list of databases.

        The function first creates a dictionary self.exc_codes to store the codes of exchanges using their attributes
        like name, reference product, location, and unit. It then fetches the unique activities by calling
        the get_list_unique_acts function using the list of the original database and scenarios as the argument.

        It then creates a dictionary acts_ind to map indices to each activity, and another dictionary acts_ind_rev
        that maps activities to their corresponding indices. It also creates a list of scenarios and a list of databases.

        The function then initializes matrices using the lil_matrix function from the scipy.sparse module to store
        the matrices for each scenario. It then extracts metadata from the databases, and stores them in a dictionary dict_meta.

        Next, for each dataset in each database, the function retrieves the dataset_id and exc_id by using their
        attributes like name, reference product, location, unit, and type. It then updates the corresponding matrix entry with the exchange amount for the current dataset and exchange.

        The function then stacks the matrices into a sparse matrix and retrieves the indices of nonzero elements.
        It then creates a new database by combining the metadata with exchanges corresponding to the indices retrieved.

        Finally, the function creates a dataframe containing the differences between scenarios and returns it along
        with the new database and a list of activities.

        :param origin_db: the original database
        :param scenarios: list of databases
        :return: a tuple containing a dataframe, a list of dictionaries, and a list of tuples
        """

        self.exc_codes.update(
            {
                (
                    dataset["name"],
                    dataset["reference product"],
                    dataset["location"],
                    dataset["unit"],
                ): dataset["code"]
                for dataset in origin_db["database"]
            }
        )

        # fetch a list of unique datasets
        list_acts = self.get_list_unique_acts(
            [origin_db["database"]] + [a["database"] for a in scenarios]
        )

        acts_ind = dict(enumerate(list_acts))
        acts_ind_rev = {value: key for key, value in acts_ind.items()}
        list_scenarios = ["original"] + [s["name"] for s in scenarios]
        list_of_databases = [origin_db["database"]] + [a["database"] for a in scenarios]

        matrices = {
            a: nsp.lil_matrix((len(list_acts), len(list_acts)))
            for a, _ in enumerate(list_scenarios)
        }

        # store the metadata in a dictionary
        dict_meta = {
            (
                dataset["name"],
                dataset["reference product"],
                None,
                dataset["location"],
                dataset["unit"],
            ): {
                key: values
                for key, values in dataset.items()
                if key
                not in [
                    "exchanges",
                    "code",
                    "name",
                    "reference product",
                    "location",
                    "unit",
                    "database",
                ]
            }
            for database in list_of_databases
            for dataset in database
        }

        for _db_index, database in enumerate(list_of_databases):
            for dataset in database:
                dataset_id = (
                    dataset["name"],
                    dataset.get("reference product"),
                    None,
                    dataset.get("location"),
                    dataset["unit"],
                    "production",
                )

                for exc in dataset["exchanges"]:
                    exc_id = (
                        self.rev_dependency_mapping.get(exc.get("input"), (None,))[0],
                        self.rev_dependency_mapping.get(exc.get("input"), (None, None))[
                            1
                        ],
                        self.rev_dependency_mapping.get(exc.get("input"), (None,))[-1],
                        self.rev_dependency_mapping.get(
                            exc.get("input"), (None, None, None)
                        )[2],
                        exc["unit"],
                        exc["type"],
                    )

                    matrices[_db_index][
                        acts_ind_rev[exc_id], acts_ind_rev[dataset_id]
                    ] += exc["amount"]

        sparse_matrix = sparse.stack(
            [sparse.COO(x) for x in matrices.values()], axis=-1
        )
        indices = sparse.argwhere(sparse_matrix.sum(-1).T != 0)
        indices = list(map(tuple, indices))

        dataframe_rows = []

        inds_d = defaultdict(list)
        for index in indices:
            inds_d[index[0]].append(index[1])

        new_db = []

        for key, value in inds_d.items():
            act = self.get_act_dict_structure(
                key,
                acts_ind,
            )
            meta_id = tuple(list(acts_ind[key])[:-1])
            act.update(dict_meta[meta_id])

            act["exchanges"].extend(
                self.get_exchange(i, acts_ind, amount=sparse_matrix[i, key, 0])
                for i in value
            )

            new_db.append(act)

        inds_std = sparse.argwhere(
            (sparse_matrix[..., 1:] == sparse_matrix[..., 0, None]).all(axis=-1).T
            == False
        )

        for _db_index in inds_std:
            c_name, c_ref, c_cat, c_loc, c_unit, c_type = acts_ind[_db_index[0]]
            s_name, s_ref, s_cat, s_loc, s_unit, s_type = acts_ind[_db_index[1]]

            if s_type == "biosphere":
                database_name = "biosphere3"

                exc_key_supplier = self.dependency_mapping[
                    (s_name, s_ref, s_loc, s_unit, s_cat)
                ]

            else:
                database_name = self.datapackage_name
                exc_key_supplier = (
                    database_name,
                    self.fetch_exchange_code(s_name, s_ref, s_loc, s_unit),
                )

            exc_key_consumer = (
                self.datapackage_name,
                self.fetch_exchange_code(c_name, c_ref, c_loc, c_unit),
            )

            row = [
                s_name,
                s_ref,
                s_loc,
                s_cat,
                database_name,
                exc_key_supplier,
                s_unit,
                c_name,
                c_ref,
                c_loc,
                c_cat,
                self.datapackage_name,
                exc_key_consumer,
                c_unit,
                s_type,
            ]

            row.extend(sparse_matrix[_db_index[1], _db_index[0], :])

            dataframe_rows.append(row)

        columns = [
            "from activity name",
            "from reference product",
            "from location",
            "from categories",
            "from database",
            "from key",
            "from unit",
            "to activity name",
            "to reference product",
            "to location",
            "to categories",
            "to database",
            "to key",
            "to unit",
            "flow type",
        ] + list_scenarios

        df = pd.DataFrame(dataframe_rows, columns=columns)
        df["to categories"] = None
        df.loc[df["flow type"] == "biosphere", "from reference product"] = None
        df.loc[df["flow type"] == "biosphere", "from location"] = None
        df.loc[df["flow type"] == "technosphere", "from categories"] = None
        df.loc[df["flow type"] == "production", "from categories"] = None
        # remove production exchanges
        df = df[df["flow type"] != "production"]

        # return the dataframe and the new db
        return df, new_db, list_acts



import pytest
from datapackage.exceptions import DataPackageException

from unfold import Unfold


def test_unfold():
    with pytest.raises(TypeError) as wrapped_error:
        u = Unfold()
    assert wrapped_error.type == TypeError

    with pytest.raises(DataPackageException) as wrapped_error:
        u = Unfold(".")
    assert wrapped_error.type == DataPackageException
